import aiomeasures
import asyncio
import pytest
from datetime import timedelta

simple = [
    (aiomeasures.CountingMetric('foo', 1), 'foo:1|c'),
    (aiomeasures.CountingMetric('foo', -1), 'foo:-1|c'),
    (aiomeasures.GaugeMetric('foo', 1), 'foo:1|g'),
    (aiomeasures.GaugeMetric('foo', -1), 'foo:-1|g'),
    (aiomeasures.HistogramMetric('foo', 42), 'foo:42|h'),
    (aiomeasures.HistogramMetric('foo', -42), 'foo:-42|h'),
    (aiomeasures.SetMetric('foo', 'bar'), 'foo:bar|s'),
    (aiomeasures.TimingMetric('foo', 100), 'foo:100|ms'),
]

one_second = timedelta(seconds=1)
twenty_ms = timedelta(microseconds=20000)

rated = [
    (aiomeasures.CountingMetric('foo', 1, 0.1), 'foo:1|c|@0.1'),
    (aiomeasures.GaugeMetric('foo', 1, 0.1), 'foo:1|g|@0.1'),
    (aiomeasures.SetMetric('foo', 'bar', 0.1), 'foo:bar|s|@0.1'),
    (aiomeasures.TimingMetric('foo', 100, 0.1), 'foo:100|ms|@0.1'),
    (aiomeasures.HistogramMetric('foo', -42, 0.1), 'foo:-42|h|@0.1'),

    (aiomeasures.CountingMetric('foo', 1, one_second), 'foo:1|c|@1'),
    (aiomeasures.GaugeMetric('foo', 1, one_second), 'foo:1|g|@1'),
    (aiomeasures.SetMetric('foo', 'bar', one_second), 'foo:bar|s|@1'),
    (aiomeasures.TimingMetric('foo', 100, one_second), 'foo:100|ms|@1'),
    (aiomeasures.HistogramMetric('foo', -42, one_second), 'foo:-42|h|@1'),

    (aiomeasures.CountingMetric('foo', 1, twenty_ms), 'foo:1|c|@0.02'),
    (aiomeasures.GaugeMetric('foo', 1, twenty_ms), 'foo:1|g|@0.02'),
    (aiomeasures.SetMetric('foo', 'bar', twenty_ms), 'foo:bar|s|@0.02'),
    (aiomeasures.TimingMetric('foo', 100, twenty_ms), 'foo:100|ms|@0.02'),
    (aiomeasures.HistogramMetric('foo', -42, twenty_ms), 'foo:-42|h|@0.02'),
]

tagged = [
    (aiomeasures.CountingMetric('foo', 1, tags={'bar': 'baz'}), 'foo:1|c|#bar:baz'),
    (aiomeasures.CountingMetric('foo', -1, tags={'bar': 'baz'}), 'foo:-1|c|#bar:baz'),
    (aiomeasures.GaugeMetric('foo', 1, tags={'bar': 'baz'}), 'foo:1|g|#bar:baz'),
    (aiomeasures.GaugeMetric('foo', -1, tags={'bar': 'baz'}), 'foo:-1|g|#bar:baz'),
    (aiomeasures.SetMetric('foo', 'bar', tags={'bar': 'baz'}), 'foo:bar|s|#bar:baz'),
    (aiomeasures.TimingMetric('foo', 100, tags={'bar': 'baz'}), 'foo:100|ms|#bar:baz'),
]

whole = simple + rated + tagged

@pytest.mark.parametrize('metric,expected', whole)
def test_formatting(metric, expected):
    handler = aiomeasures.Datadog(':0')
    assert handler.format(metric) == expected


tags = [
    ({}, {}, ''),
    ({'bar': 'baz'}, {}, '|#bar:baz'),
    ({}, {'bar': 'baz'}, '|#bar:baz'),
    ({'bar': 'baz'}, {'bar': 'qux'}, '|#bar:baz,bar:qux'),
]

@pytest.mark.parametrize('tags,defaults,expected', tags)
def test_tags(tags, defaults, expected):
    metric = aiomeasures.CountingMetric('foo', 1, tags=tags)
    handler = aiomeasures.StatsD(':0', tags=defaults)
    assert handler.format(metric) == 'foo:1|c%s' % expected


def test_events():
    event = aiomeasures.Event('Man down!', 'This server needs assistance.')
    handler = aiomeasures.StatsD(':0')
    assert handler.format(event) == '_e{9,29}Man down!|This server needs assistance.'

checks = [
    (aiomeasures.Check('srv', 'OK'), '_sc|srv|0'),
    (aiomeasures.Check('srv', 'warning'), '_sc|srv|1'),
    (aiomeasures.Check('srv', 'crit'), '_sc|srv|2'),
    (aiomeasures.Check('srv', 'UNKNOWN'), '_sc|srv|3'),
    (aiomeasures.Check('srv', 'OK', tags={'foo': 'bar'}, message='baz'), '_sc|srv|0|#foo:bar|m:baz'),
]

@pytest.mark.parametrize('check,expected', checks)
def test_checks(check,expected):
    handler = aiomeasures.Datadog(':0')
    assert handler.format(check) == expected


@asyncio.coroutine
def fake_server(event_loop, port=None):
    port = port or 0
    class Protocol:

        msg = []

        def connection_made(self, transport):
            self.transport = transport

        def datagram_received(self, data, addr):
            message = data.decode().strip()
            self.msg.extend(message.split())

        def connection_lost(self, *args):
            pass

    transport, protocol = yield from event_loop.create_datagram_endpoint(
        lambda: Protocol(),
        local_addr=('0.0.0.0', port)
    )
    if port == 0:
        _, port = transport.get_extra_info('sockname')
    return transport, protocol, port


@pytest.mark.asyncio
def test_client_event(event_loop):
    transport, protocol, port = yield from fake_server(event_loop)
    client = aiomeasures.Datadog('udp://127.0.0.1:%s' % port)
    asyncio.sleep(.4)
    client.event('title', 'text')
    yield from asyncio.sleep(.1)
    assert '_e{5,4}title|text' in protocol.msg
    transport.close()
    client.close()


@pytest.mark.asyncio
def test_client(event_loop):
    transport, protocol, port = yield from fake_server(event_loop)
    client = aiomeasures.Datadog('udp://127.0.0.1:%s' % port)
    asyncio.sleep(.4)
    client.incr('example.a')
    client.timing('example.b', 500)
    client.gauge('example.c', 1)
    client.set('example.d', 'bar')
    client.decr('example.e')
    client.counter('example.f', 42)
    client.histogram('example.g', 13)
    yield from asyncio.sleep(.1)
    assert 'example.a:1|c' in protocol.msg
    assert 'example.b:500|ms' in protocol.msg
    assert 'example.c:1|g' in protocol.msg
    assert 'example.d:bar|s' in protocol.msg
    assert 'example.e:-1|c' in protocol.msg
    assert 'example.f:42|c' in protocol.msg
    assert 'example.g:13|h' in protocol.msg
    transport.close()
    client.close()


@pytest.mark.asyncio
def test_reliablility(event_loop):
    transport, protocol, port = yield from fake_server(event_loop)
    client = aiomeasures.Datadog('udp://127.0.0.1:%s' % port)
    asyncio.sleep(.4)
    client.incr('example.a')
    yield from asyncio.sleep(.1)
    assert 'example.a:1|c' in protocol.msg
    transport.close()

    client.incr('example.b')
    yield from asyncio.sleep(.1)
    assert 'example.b:1|c' not in protocol.msg

    transport, protocol, port = yield from fake_server(event_loop, port)
    client.incr('example.c')
    yield from asyncio.sleep(.1)
    assert 'example.c:1|c' in protocol.msg
    transport.close()

    client.close()



# debug asyncio

import os
import os.path
import sys
import logging
import warnings
from pytest import fixture
from subprocess import Popen, PIPE
from time import sleep

os.environ['PYTHONASYNCIODEBUG'] = '1'
# logging.basicConfig(level=logging.DEBUG)
warnings.simplefilter("always")
warnings.filterwarnings('ignore', '.*sys.meta_path is empty.*')
warnings.filterwarnings('ignore', '.*deprecated.*', module='site')



__all__ = ['parse_addr']


class Address(tuple):
    """Defines what is a net address.
    """
    def __new__(cls, proto, host, port):
        return super().__new__(cls, (host, port))

    def __init__(self, proto, host, port):
        self.proto = proto
        self.host = host
        self.port = port

    def __hash__(self):
        return id((self.proto, self.host, self.port))

    def __eq__(self, other):
        return other == (self.proto, self.host, self.port)

    def __str__(self):
        return '%s://%s:%s' % (self.proto, self.host, self.port)


def parse_addr(addr, *, proto=None, host=None):
    """Parses an address;

    Returns:
        Address: the parsed address
    """
    port = None

    if isinstance(addr, Address):
        return addr

    elif isinstance(addr, str):
        if addr.startswith('udp://'):
            proto, addr = 'udp', addr[6:]
        elif addr.startswith('tcp://'):
            proto, addr = 'tcp', addr[6:]
        elif addr.startswith('unix://'):
            proto, addr = 'unix', addr[7:]
        a, _, b = addr.partition(':')
        host = a or host
        port = b or port

    elif isinstance(addr, (tuple, list)):
        # list is not good
        a, b = addr
        host = a or host
        port = b or port

    elif isinstance(addr, int):
        port = addr

    else:
        raise ValueError('bad value')

    if port is not None:
        port = int(port)
    return Address(proto, host, port)



from aiomeasures.checks import Check
from aiomeasures.events import Event
from aiomeasures.metrics import CountingMetric, GaugeMetric
from aiomeasures.metrics import HistogramMetric, SetMetric, TimingMetric
from collections.abc import Mapping
from datetime import datetime, timedelta
from decimal import Decimal
try:
    from functools import singledispatch
except:
    from singledispatch import singledispatch


@singledispatch
def format(obj, prefix=None, tags=None):
    raise ValueError('Cannot consume %r' % obj)


@format.register(Check)
def format_check(check, prefix=None, tags=None):
    response = '_sc|%s' % check.name
    if check.status in (0, 'ok', 'OK'):
        response += '|0'
    if check.status in (1, 'warn', 'warning', 'WARNING'):
        response += '|1'
    if check.status in (2, 'crit', 'critical', 'CRITICAL'):
        response += '|2'
    if check.status in (3, 'unknown', 'UNKNOWN'):
        response += '|3'

    if check.timestamp:
        response += '|d:%s' % format_timestamp(check.timestamp)
    if check.hostname:
        response += '|h:%s' % check.hostname
    if check.tags or tags:
        tags = format_tags(check.tags, tags)
        response += '|#%s' % ','.join(tags)
    if check.message:
        response += '|m:%s' % check.message
    return response


@format.register(Event)
def format_event(event, prefix=None, tags=None):
    a, b = len(event.title), len(event.text)
    response = '_e{%s,%s}%s|%s' % (a, b, event.title, event.text)
    if event.date_happened:
        response += '|d:%s' % format_timestamp(event.date_happened)
    if event.hostname:
        response += '|h:%s' % event.hostname
    if event.aggregation_key:
        response += '|k:%s' % event.aggregation_key
    if event.priority:
        response += '|p:%s' % event.priority
    if event.source_type_name:
        response += '|s:%s' % event.source_type_name
    if event.alert_type:
        response += '|t:%s' % event.alert_type

    if event.tags or tags:
        tags = format_tags(event.tags, tags)
        response += '|#%s' % ','.join(tags)
    return response


@format.register(CountingMetric)
def format_counting(metric, prefix=None, tags=None):
    name, value, suffix = format_metric(metric, prefix, tags)
    return '%s:%s|c%s' % (name, value, suffix)


@format.register(HistogramMetric)
def format_histogram(metric, prefix=None, tags=None):
    name, value, suffix = format_metric(metric, prefix, tags)
    return '%s:%s|h%s' % (name, value, suffix)


@format.register(GaugeMetric)
def format_gauge(metric, prefix=None, tags=None):
    name, value, suffix = format_metric(metric, prefix, tags)
    return '%s:%s|g%s' % (name, value, suffix)


@format.register(SetMetric)
def format_set(metric, prefix=None, tags=None):
    name, value, suffix = format_metric(metric, prefix, tags)
    return '%s:%s|s%s' % (name, value, suffix)


@format.register(TimingMetric)
def format_timing(metric, prefix=None, tags=None):
    name, value, suffix = format_metric(metric, prefix, tags)
    return '%s:%s|ms%s' % (name, value, suffix)


def format_metric(metric, prefix=None, tags=None):
    name = format_name(metric.name, prefix)
    value = format_value(metric.value, metric.delta)

    suffix = ''
    if metric.rate is not None:
        suffix += '|%s' % format_rate(metric.rate)

    if metric.tags or tags:
        tags = format_tags(metric.tags, tags)
        suffix += '|#%s' % ','.join(tags)

    return name, value, suffix


def format_rate(obj):
    if isinstance(obj, (float, int, Decimal)):
        return '@%s' % obj
    if isinstance(obj, timedelta):
        interval = float(obj.seconds)
        if obj.microseconds:
            interval += obj.microseconds / 1000000

        if obj.days:
            interval += obj.days * 24 * 3600
        data = '@%.3f' % interval
        data = data.rstrip('0')
        data = data.rstrip('.')
        return data


def format_tags(obj, defaults=None):
    result = set()
    for src in (obj, defaults):
        if isinstance(src, Mapping):
            result.update(['%s:%s' % (k, v) for k, v in src.items()])
        elif isinstance(src, list):
            result.update(src)
        elif isinstance(src, str):
            result.add(src)
    return sorted(result)


def format_name(name, prefix=None):
    if prefix:
        return '%s.%s' % (prefix, name)
    return '%s' % name


def format_value(value, delta=None):
    if delta and value > 0:
        return '+%s' % value
    return '%s' % value


def format_timestamp(value):
    if isinstance(value, datetime):
        return int(value.timestamp())
    if isinstance(value, float):
        return int(value)
    return value



import asyncio
import logging
from . import formatting
from aiomeasures.clients.bases import Client
from aiomeasures.collectors import Collector
from aiomeasures.reporters import StatsDReporter
from random import random


class Datadog(Client):

    def __init__(self, addr, *, prefix=None, tags=None, loop=None):
        """Sends statistics to the stats daemon over UDP

        Parameters:
            addr (str): the address in the form udp://host:port
            loop (EventLoop): the event loop
            prefix (str): prefix for all keys
            tags (dict): default tags for everything
        """
        self.loop = loop or asyncio.get_event_loop()
        self.log = logging.getLogger(__name__)
        self.prefix = prefix
        self.tags = tags
        self.collector = Collector([], 5000)
        self.reporter = StatsDReporter(addr, loop=self.loop)

    def register(self, metric):
        self.collector.append(metric)
        asyncio.Task(self.send(), loop=self.loop)
        # self.loop.async(self.send())
        return metric

    def format(self, obj):
        return formatting.format(obj, self.prefix, self.tags)

    @asyncio.coroutine
    def send(self):
        """Sends key/value pairs to Datadog agent.
        """
        yield from self.reporter.connect()
        rate = random()
        formatter = self.format
        metrics = self.collector.flush(rate=rate, formatter=formatter)
        yield from self.reporter.send(metrics)

    def close(self):
        self.reporter.close()



"""
    We’ll walk through the types of metrics supported by DogStatsD in Python,
    but the principles are easily translated into other languages.
"""

__all__ = [
    'Metric', 'CountingMetric', 'GaugeMetric',
    'HistogramMetric', 'SetMetric', 'TimingMetric'
]


class Metric:

    __slots__ = ('name', 'value', 'rate', 'delta', 'tags')

    def __init__(self, name, value, rate=None, delta=False, tags=None):
        """
        Parameters:
            name (str): name of the metric
            value (obj): value of the metric
            rate (float): sample of metric (optional)
        """

        self.name = name
        self.value = value
        self.rate = rate
        self.delta = delta
        self.tags = tags

    def __eq__(self, other):
        if isinstance(other, Metric):
            other = other.__str__()
        return self.__str__() == other

    def __repr__(self):
        args = ['%s=%s' for attr in ()]
        for attr in ('name', 'value', 'rate', 'delta', 'tags'):
            value = getattr(self, attr, None)
            if value is not None:
                args.append('%s=%r' % (attr, value))
        return '<%s(%s)>' % (self.__class__.__name__, ', '.join(args))


class CountingMetric(Metric):
    """Count things.

    Counters track how many times something happened per second,
    like the number of database requests or page views.
    """


class GaugeMetric(Metric):
    """Measure the value of a particular thing over time.

    Gauges measure the value of a particular thing at a particular time,
    like the amount of fuel in a car’s gas tank or the number of users
    connected to a system.
    """


class HistogramMetric(Metric):
    """Measure the statistical distribution of a set of values.

    Histograms track the statistical distribution of a set of values,
    like the duration of a number of database queries or the size of files
    uploaded by users.
    """


class SetMetric(Metric):
    """Count the number of unique elements in a group.

    Sets are used to count the number of unique elements in a group.
    If you want to track the number of unique visitor to your site,
    sets are a great way to do that.
    """


class TimingMetric(Metric):
    """Measure the statistical distribution of a set of values.

    StatsD only supports histograms for timing, not generic values (like
    the size of uploaded files or the number of rows returned from a query).
    Timers are essentially a special case of histograms, so they are treated
    in the same manner by DogStatsD for backwards compatibility.
    """



from collections import deque
from aiomeasures.events import Event


class Collector(deque):
    """Caped list of metrics
    """

    def flush(self, rate=None, formatter=None):
        while True:
            try:
                metric = self.popleft()
                if isinstance(metric, Event):
                    yield formatter(metric)
                    continue
                if metric.value is None:
                    continue
                if rate and metric.rate and rate < metric.rate:
                    continue
                if formatter:
                    try:
                        yield formatter(metric)
                    except ValueError:
                        continue
                else:
                    yield metric
            except IndexError:
                raise StopIteration




__all__ = ['Event']


class Event:
    """
    You can post events to your Datadog event stream.
    You can tag them, set priority and even aggregate them with other events.

    Parameters:
        title (str): Event title
        text (str): Event text. Supports line breaks

        date_happened (datetime): Assign a timestamp to the event.
                                  Default is now when none.
        hostname (str): Assign a hostname to the event.
        aggregation_key (str): Assign an aggregation key to the event,
                               to group it with some others.
        priority (str): Can be `normal` or `low`. Default to `normal`
        source_type_name (str): Assign a source type to the event.
        alert_type (str): Can be `error`, `warning`, `info` or `success`.
                          default to `info`
        tags (list): An array of tags
    """

    __slots__ = ('title', 'text', 'date_happened', 'hostname', 'alert_type',
                 'aggregation_key', 'priority', 'source_type_name', 'tags')

    def __init__(self,
                 title,
                 text,
                 date_happened=None,
                 hostname=None,
                 alert_type=None,
                 aggregation_key=None,
                 priority=None,
                 source_type_name=None,
                 tags=None):
        self.title = title
        self.text = text
        self.date_happened = date_happened
        self.hostname = hostname
        self.alert_type = alert_type
        self.aggregation_key = aggregation_key
        self.priority = priority
        self.source_type_name = source_type_name
        self.tags = tags

    def __repr__(self):
        args = ['%s=%s' for attr in ()]
        attrs = ('title', 'text', 'alert_type',
                 'aggregation_key', 'priority', 'tags')
        for attr in attrs:
            value = getattr(self, attr, None)
            if value is not None:
                args.append('%s=%r' % (attr, value))
        return '<%s(%s)>' % (self.__class__.__name__, ', '.join(args))
